import itertools

from django.conf import settings
from django.contrib.gis.geoip2 import GeoIP2
from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from geoip2.errors import AddressNotFoundError

from . import Access
from . import Settings
from . import continents_codes


class ForbidLocationMiddleware:
    """Checks if the user location is forbidden."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        city = dict()
        geoip = GeoIP2()
        address = request.META.get("REMOTE_ADDR")
        address = request.META.get("HTTP_X_FORWARDED_FOR", address)
        client_ip = address.split(",")[0].strip()
        verified_ip = request.session.get("VERIFIED_IP", "")

        # Skips if user IP is verified.
        if verified_ip == client_ip:
            return self.get_response(request)

        try:
            city = geoip.city(client_ip)

            countries = Settings.get("COUNTRIES", [])
            territories = Settings.get("TERRITORIES", [])
            country_state_code = city.get("region")
            country_identifier = city.get("country_code")

            if territories:
                # Adds the continent codes of the countries that are forbidden partially.
                territories = list({*territories, *itertools.chain.from_iterable(
                    map(continents_codes.__getitem__, filter(bool, map(Access.getattr, countries)))
                )})

            if country_state_code:
                country_identifier += ":%s" % country_state_code
            granted = all([
                Access(countries).grants(country_identifier),
                Access(territories).grants(city.get("continent_code")),
            ])
        except (AddressNotFoundError, Exception):
            # This happens when the IP address is not
            # in  the  GeoIP2 database. Usually, this
            # happens when the IP address is a local.
            granted = not any([
                Settings.has("COUNTRIES"),
                Settings.has("TERRITORIES"),
            ]) or getattr(settings, "DEBUG", False)
        finally:
            # Saves the timezone in the session for
            # comparing it with the timezone in the
            # POST request sent from user's browser
            # to detect if the user is using a VPN.
            timezone = city.get("time_zone", "N/A")
            request.session["GEOIP2_TZ"] = timezone

        if granted:
            request.session["VERIFIED_IP"] = client_ip
            return self.get_response(request)

        # Erases the timezone from the session.
        request.session["VERIFIED_IP"] = ""

        # Redirects to the FORBIDDEN_LOC URL if set.
        if Settings.has("OPTIONS.URL.FORBIDDEN_LOC"):
            return redirect(Settings.get("OPTIONS.URL.FORBIDDEN_LOC"))

        return HttpResponseForbidden()



import json

from django.http import HttpResponse
from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.shortcuts import render

from . import Settings


class ForbidNetworkMiddleware:
    """Checks if the user network is forbidden."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response_attributes = ("content", "charset", "status", "reason")

        def erase_response_attributes():
            for attr in response_attributes:
                request.session.pop(attr)

        def forbidden_page():
            # Redirects to the FORBIDDEN_NET URL if set.
            if Settings.has("OPTIONS.URL.FORBIDDEN_NET"):
                return redirect(Settings.get("OPTIONS.URL.FORBIDDEN_NET"))
            return HttpResponseForbidden()

        geoip2_tz = request.session.get("GEOIP2_TZ")
        verified_tz = request.session.get("VERIFIED_TZ", "")

        # Checks if the user's timezone match with the last accessed one.
        if verified_tz == geoip2_tz or not Settings.get("OPTIONS.VPN", False):
            return self.get_response(request)
        # Checks if GEOIP2_TZ and VERIFIED_TZ don't exist.
        elif verified_tz and geoip2_tz != "N/A":
            return forbidden_page()

        if all(map(request.session.has_key, ("GEOIP2_TZ", *response_attributes))):
            # Handles if the user's timezone differs from the
            # one determined by GeoIP API. If so, VPN is used.
            client_tz = request.POST.get("CLIENT_TZ", verified_tz)

            if geoip2_tz != "N/A" and client_tz != geoip2_tz:
                request.session["VERIFIED_TZ"] = ""
                erase_response_attributes()
                return forbidden_page()

            # Restores the response from the session.
            response = HttpResponse(**{
                attr: request.session.get(attr) for attr in response_attributes
            })
            if hasattr(response, "headers"):
                response.headers = json.loads(request.session.get("headers"))
            request.session["VERIFIED_TZ"] = geoip2_tz
            erase_response_attributes()
            return response

        # Gets the response and saves attributes in the session to restore it later.
        response = self.get_response(request)
        if hasattr(response, "headers"):
            # In older versions of Django, HttpResponse does not have headers.
            request.session["headers"] = json.dumps(dict(response.headers))
        request.session["content"] = response.content.decode(response.charset)
        request.session["charset"] = response.charset
        request.session["status"] = response.status_code
        request.session["reason"] = response.reason_phrase

        return render(request, "timezone.html", status=302)



from device_detector import DeviceDetector
from django.http import HttpResponseForbidden
from django.shortcuts import redirect

from . import Access
from . import Settings


class ForbidDeviceMiddleware:
    """Checks if the user device is forbidden."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        device_aliases = {
            "portable media player": "player",
            "smart display": "display",
            "smart speaker": "speaker",
            "feature phone": "phone",
            "car browser": "car",
        }

        devices = Settings.get("DEVICES", [])
        device_type = request.session.get("DEVICE")
        http_ua = request.META.get("HTTP_USER_AGENT")
        verified_ua = request.session.get("VERIFIED_UA", "")

        # Skips if DEVICES empty or user agent is verified.
        if not devices or verified_ua == http_ua:
            return self.get_response(request)

        if not device_type:
            device_detector = DeviceDetector(http_ua)
            device_detector = device_detector.parse()
            device = device_detector.device_type()
            device_type = device_aliases.get(device, device)
            request.session["DEVICE"] = device_type

        if Access(devices).grants(device_type):
            request.session["VERIFIED_UA"] = http_ua
            return self.get_response(request)

        # Erases the user agent from the session.
        request.session["VERIFIED_UA"] = ""

        # Redirects to the FORBIDDEN_DEV URL if set.
        if Settings.has("OPTIONS.URL.FORBIDDEN_DEV"):
            return redirect(Settings.get("OPTIONS.URL.FORBIDDEN_DEV"))

        return HttpResponseForbidden()



import re

from django.conf import settings


class Settings:
    """A helper class to access settings in a more convenient way."""

    @classmethod
    def _get(cls, item):
        result = getattr(settings, "DJANGO_FORBID", {})
        for attr in item.split("."):
            result = result[attr]
        return result

    @classmethod
    def has(cls, item):
        try:
            cls._get(item)
            return True
        except KeyError:
            return False

    @classmethod
    def get(cls, item, default=None):
        try:
            return cls._get(item)
        except KeyError:
            return default


class Access:
    """A helper class to check if a user has access to a resource."""

    def __init__(self, attributes):
        self.attributes = attributes

    @staticmethod
    def normalize(attribute_type):
        """Removes the "!" prefix."""
        return attribute_type[1:]

    @staticmethod
    def forbidden(attribute_type):
        """Checks if the type is forbidden."""
        return attribute_type.startswith("!")

    @classmethod
    def permitted(cls, attribute_type):
        """Checks if the type is permitted."""
        return not cls.forbidden(attribute_type)

    @staticmethod
    def getattr(attribute_type):
        """Checks if the type has an attribute and parses it."""
        matching_attribute = re.match(r"^!(\w+):\w+$", attribute_type)
        return matching_attribute.group(1) if matching_attribute else None

    def grants(self, attribute_type):
        # Creates a regular expression in the following form:
        # ^(?=PERMITTED_ATTRIBUTES)(?:(?!FORBIDDEN_ATTRIBUTES)\w+(?::\w+)?)$
        # where the list of forbidden and permitted attributes is determined
        # by filtering the particular setting attributes by the "!" prefix.
        permit = r"|".join(filter(self.permitted, self.attributes))
        permit = r"\w+" if any(filter(self.getattr, self.attributes)) else permit
        forbid = r"|".join(map(self.normalize, filter(self.forbidden, self.attributes)))
        forbid = r"(?!" + forbid + r")" if forbid else ""
        regexp = r"^(?=" + permit + r")(?:" + forbid + r"\w+(?::\w+)?)$"

        # Regexp designed to match the permitted attributes.
        return re.match(regexp, attribute_type)


continents_codes = {
    'AF': ['AS'], 'AL': ['EU'], 'AQ': ['AN'], 'DZ': ['AF'], 'AS': ['OC'], 'AD': ['EU'], 'AO': ['AF'], 'AG': ['NA'],
    'AZ': ['EU', 'AS'], 'AR': ['SA'], 'AU': ['OC'], 'AT': ['EU'], 'BS': ['NA'], 'BH': ['AS'], 'BD': ['AS'],
    'AM': ['EU', 'AS'], 'BB': ['NA'], 'BE': ['EU'], 'BM': ['NA'], 'BT': ['AS'], 'BO': ['SA'], 'BA': ['EU'],
    'BW': ['AF'], 'BV': ['AN'], 'BR': ['SA'], 'BZ': ['NA'], 'IO': ['AS'], 'SB': ['OC'], 'VG': ['NA'], 'BN': ['AS'],
    'BG': ['EU'], 'MM': ['AS'], 'BI': ['AF'], 'BY': ['EU'], 'KH': ['AS'], 'CM': ['AF'], 'CA': ['NA'], 'CV': ['AF'],
    'KY': ['NA'], 'CF': ['AF'], 'LK': ['AS'], 'TD': ['AF'], 'CL': ['SA'], 'CN': ['AS'], 'TW': ['AS'], 'CX': ['AS'],
    'CC': ['AS'], 'CO': ['SA'], 'KM': ['AF'], 'YT': ['AF'], 'CG': ['AF'], 'CD': ['AF'], 'CK': ['OC'], 'CR': ['NA'],
    'HR': ['EU'], 'CU': ['NA'], 'CY': ['EU', 'AS'], 'CZ': ['EU'], 'BJ': ['AF'], 'DK': ['EU'], 'DM': ['NA'],
    'DO': ['NA'], 'EC': ['SA'], 'SV': ['NA'], 'GQ': ['AF'], 'ET': ['AF'], 'ER': ['AF'], 'EE': ['EU'], 'FO': ['EU'],
    'FK': ['SA'], 'GS': ['AN'], 'FJ': ['OC'], 'FI': ['EU'], 'AX': ['EU'], 'FR': ['EU'], 'GF': ['SA'], 'PF': ['OC'],
    'TF': ['AN'], 'DJ': ['AF'], 'GA': ['AF'], 'GE': ['EU', 'AS'], 'GM': ['AF'], 'PS': ['AS'], 'DE': ['EU'],
    'GH': ['AF'], 'GI': ['EU'], 'KI': ['OC'], 'GR': ['EU'], 'GL': ['NA'], 'GD': ['NA'], 'GP': ['NA'], 'GU': ['OC'],
    'GT': ['NA'], 'GN': ['AF'], 'GY': ['SA'], 'HT': ['NA'], 'HM': ['AN'], 'VA': ['EU'], 'HN': ['NA'], 'HK': ['AS'],
    'HU': ['EU'], 'IS': ['EU'], 'IN': ['AS'], 'ID': ['AS'], 'IR': ['AS'], 'IQ': ['AS'], 'IE': ['EU'], 'IL': ['AS'],
    'IT': ['EU'], 'CI': ['AF'], 'JM': ['NA'], 'JP': ['AS'], 'KZ': ['EU', 'AS'], 'JO': ['AS'], 'KE': ['AF'],
    'KP': ['AS'], 'KR': ['AS'], 'KW': ['AS'], 'KG': ['AS'], 'LA': ['AS'], 'LB': ['AS'], 'LS': ['AF'], 'LV': ['EU'],
    'LR': ['AF'], 'LY': ['AF'], 'LI': ['EU'], 'LT': ['EU'], 'LU': ['EU'], 'MO': ['AS'], 'MG': ['AF'], 'MW': ['AF'],
    'MY': ['AS'], 'MV': ['AS'], 'ML': ['AF'], 'MT': ['EU'], 'MQ': ['NA'], 'MR': ['AF'], 'MU': ['AF'], 'MX': ['NA'],
    'MC': ['EU'], 'MN': ['AS'], 'MD': ['EU'], 'ME': ['EU'], 'MS': ['NA'], 'MA': ['AF'], 'MZ': ['AF'], 'OM': ['AS'],
    'NA': ['AF'], 'NR': ['OC'], 'NP': ['AS'], 'NL': ['EU'], 'AN': ['NA'], 'CW': ['NA'], 'AW': ['NA'], 'SX': ['NA'],
    'BQ': ['NA'], 'NC': ['OC'], 'VU': ['OC'], 'NZ': ['OC'], 'NI': ['NA'], 'NE': ['AF'], 'NG': ['AF'], 'NU': ['OC'],
    'NF': ['OC'], 'NO': ['EU'], 'MP': ['OC'], 'UM': ['OC', 'NA'], 'FM': ['OC'], 'MH': ['OC'], 'PW': ['OC'],
    'PK': ['AS'], 'PA': ['NA'], 'PG': ['OC'], 'PY': ['SA'], 'PE': ['SA'], 'PH': ['AS'], 'PN': ['OC'], 'PL': ['EU'],
    'PT': ['EU'], 'GW': ['AF'], 'TL': ['AS'], 'PR': ['NA'], 'QA': ['AS'], 'RE': ['AF'], 'RO': ['EU'], 'RW': ['AF'],
    'RU': ['EU', 'AS'], 'BL': ['NA'], 'SH': ['AF'], 'KN': ['NA'], 'AI': ['NA'], 'LC': ['NA'], 'MF': ['NA'],
    'PM': ['NA'], 'VC': ['NA'], 'SM': ['EU'], 'ST': ['AF'], 'SA': ['AS'], 'SN': ['AF'], 'RS': ['EU'], 'SC': ['AF'],
    'SL': ['AF'], 'SG': ['AS'], 'SK': ['EU'], 'VN': ['AS'], 'SI': ['EU'], 'SO': ['AF'], 'ZA': ['AF'], 'ZW': ['AF'],
    'ES': ['EU'], 'SS': ['AF'], 'EH': ['AF'], 'SD': ['AF'], 'SR': ['SA'], 'SJ': ['EU'], 'SZ': ['AF'], 'SE': ['EU'],
    'CH': ['EU'], 'SY': ['AS'], 'TJ': ['AS'], 'TH': ['AS'], 'TG': ['AF'], 'TK': ['OC'], 'TO': ['OC'], 'TT': ['NA'],
    'AE': ['AS'], 'TN': ['AF'], 'TR': ['EU', 'AS'], 'TM': ['AS'], 'TC': ['NA'], 'TV': ['OC'], 'UG': ['AF'],
    'UA': ['EU'], 'MK': ['EU'], 'EG': ['AF'], 'GB': ['EU'], 'GG': ['EU'], 'JE': ['EU'], 'IM': ['EU'], 'TZ': ['AF'],
    'US': ['NA'], 'VI': ['NA'], 'BF': ['AF'], 'UY': ['SA'], 'UZ': ['AS'], 'VE': ['SA'], 'WF': ['OC'], 'WS': ['OC'],
    'YE': ['AS'], 'ZM': ['AF'], 'XX': ['OC'], 'XE': ['AS'], 'XD': ['AS'], 'XS': ['AS']
}



import re

from .skills.forbid_device import ForbidDeviceMiddleware
from .skills.forbid_location import ForbidLocationMiddleware
from .skills.forbid_network import ForbidNetworkMiddleware

__skills__ = (
    ForbidNetworkMiddleware,
    ForbidLocationMiddleware,
    ForbidDeviceMiddleware,
)


class ForbidMiddleware:
    """Middleware to forbid access to the site."""

    def __init__(self, get_response):
        self.get_response = get_response
        self.regex = re.compile(r"\w+/(?:html|xhtml\+xml|xml)")

    def __call__(self, request):
        get_response = self.get_response
        if self.regex.search(request.META.get("HTTP_ACCEPT")):
            for skill in __skills__:
                get_response = skill(get_response)
        return get_response(request)
